@@ -19,56 +19,70 @@
 Graphical representation of a node on the QGraphicsScene.
 """
 
-from ..qt import QtCore, QtGui, QtSvg
-from .note_item import NoteItem
+from ..qt import sip
+
+from ..qt import QtCore, QtGui, QtWidgets, QtSvg, qslot
+from ..qt.qimage_svg_renderer import QImageSvgRenderer
+from .label_item import LabelItem
+from ..symbol import Symbol
+from ..controller import Controller
+
+
+import logging
+log = logging.getLogger(__name__)
 
 
 class NodeItem(QtSvg.QGraphicsSvgItem):
+
     """
     Node for the scene.
 
     :param node: Node instance
-    :param default_symbol: Default symbol for the node representation on the scene
-    :param hover_symbol: Hover symbol when the node is hovered on the scene
     """
 
     show_layer = False
 
-    def __init__(self, node, default_symbol=None, hover_symbol=None):
-
-        QtSvg.QGraphicsSvgItem.__init__(self)
+    def __init__(self, node):
+        super().__init__()
 
         # attached node
         self._node = node
+        # link items connected to this node item.
+        self._links = []
+        self._symbol = None
+        self._locked = False
+        self._allow_snap_to_grid = True
+
+        # says if the attached node has been initialized
+        # by the server.
+        self._initialized = False
 
         # node label
         self._node_label = None
 
-        # link items connected to this node item.
-        self._links = []
+        self.setPos(QtCore.QPoint(self._node.x(), self._node.y()))
+
+        # Temporary symbol during loading
+        renderer = QImageSvgRenderer(":/icons/reload.svg")
+        renderer.setObjectName("symbol_loading")
+        self.setSharedRenderer(renderer)
+
+        effect = QtWidgets.QGraphicsColorizeEffect()
+        effect.setColor(QtGui.QColor("black"))
+        effect.setStrength(0.8)
+        self.setGraphicsEffect(effect)
+        self.graphicsEffect().setEnabled(False)
 
         # set graphical settings for this node
-        self.setFlag(QtSvg.QGraphicsSvgItem.ItemIsMovable)
-        self.setFlag(QtSvg.QGraphicsSvgItem.ItemIsSelectable)
-        self.setFlag(QtSvg.QGraphicsSvgItem.ItemIsFocusable)
-        self.setFlag(QtSvg.QGraphicsSvgItem.ItemSendsGeometryChanges)
-        self.setAcceptsHoverEvents(True)
-        self.setZValue(1)
+        self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable)
+        self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable)
+        self.setFlag(QtWidgets.QGraphicsItem.ItemIsFocusable)
+        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsGeometryChanges)
+        self.setAcceptHoverEvents(True)
 
-        # create renderers using symbols paths/resources
-        if default_symbol:
-            self._default_renderer = QtSvg.QSvgRenderer(default_symbol)
-            if default_symbol != node.defaultSymbol():
-                self._default_renderer.setObjectName(default_symbol)
-        else:
-            self._default_renderer = QtSvg.QSvgRenderer(node.defaultSymbol())
-        if hover_symbol:
-            self._hover_renderer = QtSvg.QSvgRenderer(hover_symbol)
-            if hover_symbol != node.hoverSymbol():
-                self._hover_renderer.setObjectName(hover_symbol)
-        else:
-            self._hover_renderer = QtSvg.QSvgRenderer(node.hoverSymbol())
-        self.setSharedRenderer(self._default_renderer)
+        # update z value and locked state
+        self.setLocked(self._node.locked())
+        self.setZValue(self._node.z())
 
         # connect signals to know about some events
         # e.g. when the node has been started, stopped or suspended etc.
@@ -78,66 +92,70 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         node.suspended_signal.connect(self.suspendedSlot)
         node.updated_signal.connect(self.updatedSlot)
         node.deleted_signal.connect(self.deletedSlot)
-        node.delete_links_signal.connect(self.deleteLinksSlot)
         node.error_signal.connect(self.errorSlot)
         node.server_error_signal.connect(self.serverErrorSlot)
 
         # used when a port has been selected from the contextual menu
         self._selected_port = None
 
-        # says if the attached node has been initialized
-        # by the server.
-        self._initialized = False
-
         # contains the last error message received
         # from the server.
         self._last_error = None
 
-    def defaultRenderer(self):
-        """
-        Returns the default QSvgRenderer.
-
-        :return: QSvgRenderer instance
-        """
-
-        return self._default_renderer
-
-    def setDefaultRenderer(self, default_renderer):
-        """
-        Sets new default QSvgRenderer.
-
-        :param default_renderer: QSvgRenderer instance
-        """
-
-        self._default_renderer = default_renderer
-        self.setSharedRenderer(self._default_renderer)
-
-    def hoverRenderer(self):
-        """
-        Returns the hover QSvgRenderer.
-
-        :return: QSvgRenderer instance
-        """
-
-        return self._hover_renderer
-
-    def setHoverRenderer(self, hover_renderer):
-        """
-        Sets new hover QSvgRenderer.
-
-        :param hover_renderer: QSvgRenderer instance
-        """
-
-        self._hover_renderer = hover_renderer
-
-    def setUnsavedState(self):
-        """
-        Indicates the project is in a unsaved state.
-        """
-
         from ..main_window import MainWindow
-        main_window = MainWindow.instance()
-        main_window.setUnsavedState()
+        self._main_window = MainWindow.instance()
+        self._settings = self._main_window.uiGraphicsView.settings()
+
+        if node.initialized():
+            self.createdSlot(node.id())
+
+        if self._main_window.uiSnapToGridAction.isChecked():
+            self.setPos(QtCore.QPointF(self._node.x() + 0.1, self._node.y()))
+
+    def updateNode(self):
+        """
+        Sync change to the node
+        """
+
+        self._node.setGraphics(self)
+
+    @qslot
+    def setSymbol(self, symbol):
+        """
+        :param symbol: Change the symbol path
+        """
+        # create renderer using symbols path/resource
+        if symbol is None:
+            symbol = self._node.defaultSymbol()
+        if self._symbol != symbol:
+            self._symbol = symbol
+
+            # Temporary symbol during loading
+            renderer = QImageSvgRenderer(":/icons/reload.svg")
+            renderer.setObjectName("symbol_loading")
+            self.setSharedRenderer(renderer)
+
+            Controller.instance().getStatic(Symbol(symbol_id=symbol).url(), self._symbolLoadedCallback)
+
+    def symbol(self):
+        return self._symbol
+
+    @qslot
+    def _symbolLoadedCallback(self, path, *args):
+
+        renderer = QImageSvgRenderer(path, fallback=":/icons/cancel.svg")
+        renderer.setObjectName(path)
+        self.setSharedRenderer(renderer)
+        if self._settings["limit_size_node_symbols"] is True and renderer.defaultSize().height() > 80:
+            # resize the SVG
+            renderer.resize(80)
+            self.setSharedRenderer(renderer)
+        if self._node.settings().get("symbol") != self._symbol:
+            self.updateNode()
+        if not self._initialized:
+            self._showLabel()
+            self._initialized = True
+            self.updateNode()
 
     def node(self):
         """
@@ -148,27 +166,44 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
 
         return self._node
 
-    def addLink(self, link):
+    def setPos(self, *args):
+        super().setPos(*args)
+        self._node.setSettingValue("x", int(self.x()))
+        self._node.setSettingValue("y", int(self.y()))
+
+    @qslot
+    def addLink(self, link_item, *args):
         """
         Adds a link items to this node item.
 
         :param link: LinkItem instance
         """
 
-        self._links.append(link)
-        self._node.updated_signal.emit()
-        self.setUnsavedState()
+        if not sip.isdeleted(link_item):
+            self._links.append(link_item)
+            link_item.link().delete_link_signal.connect(self._removeLink)
+            link_item.link().updated_link_signal.connect(self._linkUpdatedSlot)
+            self._node.updated_signal.emit()
 
-    def removeLink(self, link):
+    @qslot
+    def _linkUpdatedSlot(self, *args):
+        """
+        When a link change we also notify the listener of the node
+        """
+        self._node.updated_signal.emit()
+
+    @qslot
+    def _removeLink(self, link_id, *args):
         """
         Removes a link items from this node item.
 
         :param link: LinkItem instance
         """
 
-        if link in self._links:
-            self._links.remove(link)
-        self.setUnsavedState()
+        for link_item in self._links:
+            if link_item.link().id() == link_id:
+                self._links.remove(link_item)
+                return
 
     def links(self):
         """
@@ -179,19 +214,21 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
 
         return self._links
 
-    def createdSlot(self, node_id):
+    @qslot
+    def createdSlot(self, base_node_id, *args):
         """
         Slot to receive events from the attached Node instance
         when a the node has been created/initialized.
 
-        :param node_id: node identifier (integer)
+        :param base_node_id: base node identifier (integer)
         """
 
-        self._initialized = True
+        self.setPos(QtCore.QPoint(self._node.x(), self._node.y()))
+        self.setSymbol(self._node.symbol())
         self.update()
-        self._showLabel()
 
-    def startedSlot(self):
+    @qslot
+    def startedSlot(self, *args):
         """
         Slot to receive events from the attached Node instance
         when a the node has started.
@@ -200,7 +237,8 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         for link in self._links:
             link.update()
 
-    def stoppedSlot(self):
+    @qslot
+    def stoppedSlot(self, *args):
         """
         Slot to receive events from the attached Node instance
         when a the node has stopped.
@@ -209,7 +247,8 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         for link in self._links:
             link.update()
 
-    def suspendedSlot(self):
+    @qslot
+    def suspendedSlot(self, *args):
         """
         Slot to receive events from the attached Node instance
         when a the node has suspended.
@@ -218,61 +257,55 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         for link in self._links:
             link.update()
 
-    def updatedSlot(self):
+    @qslot
+    def updatedSlot(self, *args):
         """
         Slot to receive events from the attached Node instance
         when a the node has been updated.
         """
 
-        if self._node_label:
-            if self._node_label.toPlainText() != self._node.name():
-                self._node_label.setPlainText(self._node.name())
-                self._centerLabel()
-        self.setUnsavedState()
+        self.setSymbol(self._node.settings().get("symbol"))
+        self.setPos(self._node.settings().get("x", 0), self._node.settings().get("y", 0))
+        self.setZValue(self._node.settings().get("z", 0))
+        self.setLocked(self._node.settings().get("locked", False))
+        self._updateLabel()
 
         # update the link tooltips in case the
         # node name has changed
         for link in self._links:
             link.setCustomToolTip()
 
-    def deleteLinksSlot(self):
-        """
-        Slot to receive events from the attached Node instance
-        when a all the links must be deleted.
-        """
-
-        for link in self._links.copy():
-            link.delete()
-
-    def deletedSlot(self):
+    @qslot
+    def deletedSlot(self, *args):
         """
         Slot to receive events from the attached Node instance
         when the node has been deleted.
         """
 
-        self._node.removeAllocatedName()
+        if not self.scene():
+            return
         if self in self.scene().items():
             self.scene().removeItem(self)
-        self.setUnsavedState()
 
-    def serverErrorSlot(self, node_id, code, message):
+    @qslot
+    def serverErrorSlot(self, base_node_id, message, *args):
         """
         Slot to receive events from the attached Node instance
         when the node has received an error from the server.
 
-        :param node_id: node identifier
-        :param code: error code
+        :param base_node_id: base node identifier
         :param message: error message
         """
 
         self._last_error = "{message}".format(message=message)
 
-    def errorSlot(self, node_id, message):
+    @qslot
+    def errorSlot(self, base_node_id, message, *args):
         """
         Slot to receive events from the attached Node instance
         when the node wants to report an error.
 
-        :param node_id: node identifier
+        :param base_node_id: base node identifier
         :param message: error message
         """
 
@@ -301,14 +334,11 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
 
         return self._node_label
 
-    def setLabel(self, label):
+    def _labelUnselectedSlot(self):
         """
-        Sets the node label.
-
-        :param label: NoteItem instance.
+        Called when user unselect the label
         """
-
-        self._node_label = label
+        self.updateNode()
 
     def _centerLabel(self):
         """
@@ -322,6 +352,7 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         label_x_pos = node_middle.x() - text_middle.x()
         label_y_pos = -25
         self._node_label.setPos(label_x_pos, label_y_pos)
+        return
 
     def _showLabel(self):
         """
@@ -329,12 +360,39 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         """
 
         if not self._node_label:
-            self._node_label = NoteItem(self)
+            self._node_label = LabelItem(self)
+            self._node_label.item_unselected_signal.connect(self._labelUnselectedSlot)
             self._node_label.setEditable(False)
-            self._node_label.setPlainText(self._node.name())
-            self._centerLabel()
+            self._updateLabel()
+            self._node.setSettingValue("label", self._node_label.dump())
 
-    def connectToPort(self, unavailable_ports=[]):
+    def _updateLabel(self):
+        """
+        Update the label using the information stored in the node
+        """
+        if not self._node_label:
+            return
+        self._node_label.setPlainText(self._node.name())
+        label_data = self._node.settings().get("label")
+
+        if self._node_label.toPlainText() != label_data["text"]:
+            self._node_label.setPlainText(label_data["text"])
+
+        style = label_data.get("style")
+        if style:
+            self._node_label.setStyle(style)
+        self._node_label.setRotation(label_data.get("rotation", 0))
+
+        if self._node.locked():
+            self._node_label.setFlag(self.ItemIsMovable, False)
+
+        if label_data["x"] is None:
+            self._centerLabel()
+            self.updateNode()
+        else:
+            self._node_label.setPos(label_data["x"], label_data["y"])
+
+    def connectToPort(self, pos, unavailable_ports=[]):
         """
         Shows a contextual menu for the user to choose port or auto-select one.
 
@@ -344,23 +402,25 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         """
 
         self._selected_port = None
-        menu = QtGui.QMenu()
+        menu = QtWidgets.QMenu()
         ports = self._node.ports()
         if not ports:
-            QtGui.QMessageBox.critical(self.scene().parent(), "Link", "No port available, please configure this device")
+            QtWidgets.QMessageBox.critical(self.scene().parent(), "Link", "No port available, please configure this device")
             return None
 
         # sort the ports
         ports_dict = {}
         for port in ports:
-            if port.slotNumber() is not None:
-                # multiply the slot number by 16 to make the port number unique.
-                ports_dict[(port.slotNumber() * 16) + port.portNumber()] = port
+            if port.adapterNumber() is not None:
+                # make the port number unique (special case with WICs).
+                port_number = port.portNumber()
+                if port_number >= 16:
+                    port_number *= 8
+                ports_dict[(port.adapterNumber() * 16) + port_number] = port
             elif port.portNumber()is not None:
                 ports_dict[port.portNumber()] = port
             else:
                 ports_dict[port.name()] = port
-
         try:
             ports = sorted(ports_dict.keys(), key=int)
         except ValueError:
@@ -369,6 +429,7 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         # show a contextual menu for the user to choose a port
         for port in ports:
             port_object = ports_dict[port]
+            log.debug("Node '{}' Port {} Type {}".format(self.node(), port_object.name(), type(port_object.name())))
             if port in unavailable_ports:
                 # this port cannot be chosen by the user (grayed out)
                 action = menu.addAction(QtGui.QIcon(':/icons/led_green.svg'), port_object.name())
@@ -379,7 +440,10 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
                 menu.addAction(QtGui.QIcon(':/icons/led_green.svg'), port_object.name())
 
         menu.triggered.connect(self.selectedPortSlot)
-        menu.exec_(QtGui.QCursor.pos())
+        # add some delay before showing the menu
+        # https://github.com/GNS3/gns3-gui/issues/3169
+        QtCore.QThread.msleep(100)
+        menu.exec_(pos)
         return self._selected_port
 
     def selectedPortSlot(self, action):
@@ -406,20 +470,28 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         :param value: value of the change
         """
 
+        if change == QtWidgets.QGraphicsItem.ItemPositionChange and self._main_window.uiSnapToGridAction.isChecked() \
+                and self._allow_snap_to_grid:
+            grid_size = self._main_window.uiGraphicsView.nodeGridSize()
+            mid_x = self.boundingRect().width() / 2
+            value.setX((grid_size * round((value.x() + mid_x) / grid_size)) - mid_x)
+            mid_y = self.boundingRect().height() / 2
+            value.setY((grid_size * round((value.y() + mid_y) / grid_size)) - mid_y)
+
         # dynamically change the renderer when this node item is selected/unselected.
-        if change == QtSvg.QGraphicsSvgItem.ItemSelectedChange:
+        if change == QtWidgets.QGraphicsItem.ItemSelectedChange:
             if value:
-                self.setSharedRenderer(self._hover_renderer)
+                self.graphicsEffect().setEnabled(True)
             else:
-                self.setSharedRenderer(self._default_renderer)
+                self.graphicsEffect().setEnabled(False)
+                self.updateNode()
 
         # adjust link item positions when this node is moving or has changed.
-        if change == QtSvg.QGraphicsSvgItem.ItemPositionChange or change == QtSvg.QGraphicsSvgItem.ItemPositionHasChanged:
-            self.setUnsavedState()
+        if change == QtWidgets.QGraphicsItem.ItemPositionChange or change == QtWidgets.QGraphicsItem.ItemPositionHasChanged:
             for link in self._links:
                 link.adjust()
 
-        return QtGui.QGraphicsItem.itemChange(self, change, value)
+        return super().itemChange(change, value)
 
     def paint(self, painter, option, widget=None):
         """
@@ -431,15 +503,16 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         """
 
         # don't show the selection rectangle
-        option.state = QtGui.QStyle.State_None
-        QtSvg.QGraphicsSvgItem.paint(self, painter, option, widget)
+        if not self._settings["draw_rectangle_selected_item"]:
+            option.state = QtWidgets.QStyle.State_None
+        super().paint(painter, option, widget)
 
         if not self._initialized or self.show_layer:
             brect = self.boundingRect()
             center = self.mapFromItem(self, brect.width() / 2.0, brect.height() / 2.0)
             painter.setBrush(QtCore.Qt.red)
             painter.setPen(QtCore.Qt.red)
-            painter.drawRect((brect.width() / 2.0) - 10, (brect.height() / 2.0) - 10, 20, 20)
+            painter.drawRect(QtCore.QRectF((brect.width() / 2.0) - 10, (brect.height() / 2.0) - 10, 20, 20))
             painter.setPen(QtCore.Qt.black)
             if self.show_layer:
                 text = str(int(self.zValue()))  # Z value
@@ -456,21 +529,53 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         :param value: Z value
         """
 
-        QtSvg.QGraphicsSvgItem.setZValue(self, value)
-        if self.zValue() < 0:
-            self.setFlag(self.ItemIsSelectable, False)
+        super().setZValue(value)
+        for link in self._links:
+            link.adjust()
+
+    def keyPressEvent(self, event):
+        """
+        Handles all key press events
+
+        :param event: QKeyEvent
+        """
+
+        if event.modifiers() & QtCore.Qt.AltModifier:
+            self._allow_snap_to_grid = False
+        else:
+            super().keyPressEvent(event)
+
+    def keyReleaseEvent(self, event):
+        """
+        Handles all key release events
+
+        :param event: QKeyEvent
+        """
+
+        self._allow_snap_to_grid = True
+
+    def locked(self):
+
+        return self._locked
+
+    def setLocked(self, locked):
+        """
+        Sets the locked value.
+
+        :param value: Z value
+        """
+
+        if locked is True:
             self.setFlag(self.ItemIsMovable, False)
             if self._node_label:
-                self._node_label.setFlag(self.ItemIsSelectable, False)
                 self._node_label.setFlag(self.ItemIsMovable, False)
         else:
-            self.setFlag(self.ItemIsSelectable, True)
             self.setFlag(self.ItemIsMovable, True)
             if self._node_label:
-                self._node_label.setFlag(self.ItemIsSelectable, True)
                 self._node_label.setFlag(self.ItemIsMovable, True)
         for link in self._links:
             link.adjust()
+        self._locked = locked
 
     def hoverEnterEvent(self, event):
         """
@@ -480,13 +585,8 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         """
 
         self.setCustomToolTip()
-        # dynamically change the renderer when this node item is hovered.
         if not self.isSelected():
-            self.setSharedRenderer(self._hover_renderer)
-            #effect = QtGui.QGraphicsColorizeEffect()
-            #effect.setColor(QtGui.QColor("black"))
-            #effect.setStrength(0.8)
-            #self.setGraphicsEffect(effect)
+            self.graphicsEffect().setEnabled(True)
 
     def hoverLeaveEvent(self, event):
         """
@@ -495,7 +595,13 @@ class NodeItem(QtSvg.QGraphicsSvgItem):
         :param event: QGraphicsSceneHoverEvent instance
         """
 
-        # dynamically change the renderer back to the default when this node item is not hovered anymore.
         if not self.isSelected():
-            self.setSharedRenderer(self._default_renderer)
-            #self.graphicsEffect().setEnabled(False)
+            self.graphicsEffect().setEnabled(False)
+
+    def mouseRelease(self):
+        """
+        Handle all mouse release for this item.
+        It the item is select but mouse is not on it the event
+        is send also
+        """
+        self.updateNode()
